/*    Filename: pthread_kill_latency.c
 *      Author: Sripathi Kodi <sripathik@in.ibm.com>
 * Description: Measure the latency involved in sending a signal to a thread
 * using pthread_kill. Two threads are created - the one that receives the
 * signal (thread1) and the other that sends the signal (thread2). Before
 * sending the signal, the thread2 waits for thread1 to initialize, notes the
 * time and sends pthread_kill signal to thread1. thread2, which has defined a
 * handler for the signal, notes the time it recieves the signal. The maximum
 * and the minimum latency is reported.
 *
 * Compilation: gcc pthread_kill_latency.c -lpthread -lrt -lm -D_GNU_SOURCE
 * -I/usr/include/nptl -o pthread_kill_latency
 *
 * Usage      : pthread_kill_latency [-v{1234}]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Copyright (C) IBM Corporation, 2006
 * 2006-Jun-28:  Initial version by Sripathi Kodi <sripathik@in.ibm.com>
 */

#include <stdio.h>
#include <time.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <errno.h>
#include "librt.h"
#include "libjvmsim.h"

#define PRIO 89
#define NUMLOOPS 10000
#define THRESHOLD 20000
#define SIGNALNUMBER SIGUSR1

/* Get the pthread structure corresponding to this thread id */
#define PTHREADOF(tid) _threads[(tid)].pthread

nsec_t this, begin, end, max, min;
int fail;
float avg;

atomic_t flag;

#if 0
/* Set up a signal handler */
int rt_setsighandler(int signum, void(*handler)(int))
{
	struct sigaction sa;
	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = handler;
	if (sigaction(signum, &sa, NULL) != 0) {
		perror("Sigaction failed:\n");
		return 1;
	}
	return 0;
}
#endif

void *signal_receiving_thread(void *arg)
{
	int i, ret, sig; 
	sigset_t set, oset;

	debug(DBG_DEBUG, "Signal receiving thread running\n");

	if ((sigaddset(&set, SIGNALNUMBER))) {
		perror("sigaddset:");
		exit(1);
	}
	if ((ret = pthread_sigmask(SIG_BLOCK, &set, &oset))) {
		printf("pthread_sigmask returned %d\n", ret);
		exit(1);
	}

	/* Let the sending thread know that receiver is ready */
	atomic_set(1, &flag);

	debug(DBG_DEBUG, "Signal receiving thread ready to receive\n");

	/* Warm up */
	for (i=0; i<5; i++) {
		sigwait(&set, &sig);
		atomic_set(1,&flag);
	}

	max=0; min=-1; avg=0; fail=0;
	printf("\n\n");
	for (i=0; i<NUMLOOPS; i++) {
		sigwait(&set, &sig);
		end = rt_gettime();
		this = end - begin;
		if (this > max)
			max = this;
		if (this < min)
			min = this;
		avg += (this/NS_PER_US);
		if (this > THRESHOLD) fail++;

		debug(DBG_INFO, "Iteration %d: Took %Ld ns. Max = %Ld, Min = %Ld\n", i, this, max, min);
		printf("Iteration: %d, Max time: %Ld, Min time: %Ld, Failures: %d\r", i, max, min, fail);
		fflush(stdout);
		buffer_print();
		atomic_set(1,&flag);
	}
	printf("\n\n");
	printf("Average time spent = %4.2f microseconds\n", avg/NUMLOOPS);
	printf("\n\n");
	return NULL;
}

void *signal_sending_thread(void *arg)
{
	int target_thread = (int)((struct thread *)arg)->arg;
	int i, ret;

	debug(DBG_INFO, "Signal sending thread: target thread id =%d\n", (int)PTHREADOF(target_thread));

	/* Wait for the receiving thread to initialize */
	while(!atomic_get(&flag)) {usleep(100);};
	atomic_set(0, &flag);

	/* Warm up */
	for (i=0; i<5; i++) {

		debug(DBG_DEBUG, "Sending signal (Warm up). Loopcnt = %d\n", i);

		if ((ret = pthread_kill(PTHREADOF(target_thread), SIGNALNUMBER))) {
			printf("pthread_kill returned %d\n", ret);
		}
		/* Wait till the receiving thread processes the signal */
		while(!atomic_get(&flag)) {usleep(100);};
		atomic_set(0, &flag);
	}
	for (i=0; i<NUMLOOPS; i++) {

		debug(DBG_DEBUG, "Sending signal. Loopcnt = %d\n", i);

		/* Record the time just before sending the signal */
		begin = rt_gettime();
		if ((ret = pthread_kill(PTHREADOF(target_thread), SIGNALNUMBER))) {
			printf("pthread_kill returned %d\n", ret);
		}
		/* Wait till the receiving thread processes the signal */
		while(!atomic_get(&flag)) {usleep(100);};
		atomic_set(0, &flag);
	}
	return NULL;
}
	
int main(int argc, char *argv[])
{
	int thr_id1, thr_id2;

	atomic_set(0,&flag);

//	jvmsim_init();
	rt_init(NULL, NULL, argc, argv);	/* we need the buffered print system */
	
	debug(DBG_DEBUG, "Main creating threads\n");
	thr_id1 = create_fifo_thread(signal_receiving_thread, (void*)0, PRIO);
//	thr_id2 = create_fifo_thread(signal_sending_thread, (void*)thr_id1, PRIO);
	thr_id2 = create_other_thread(signal_sending_thread, (void*)thr_id1);

	debug(DBG_DEBUG, "Main joining threads debug\n");
	join_thread(thr_id1);
	join_thread(thr_id2);
	buffer_print();
	return 0;
}
